/*
* avrapp.c
*
* Created: 19/10/2012 13:11:55
*  Author: cliff
*/

#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#if (F_CPU != 32000000UL)
#error F_CPU must match oscillator setting!
#endif // F_CPU

#define DEVID  0x12345678
#define SWVERSIONMAJOR 0
#define SWVERSIONMINOR 0

#include "usart_driver.h"
USART_data_t USART_data;

typedef struct
{
	unsigned long dev_id;
	unsigned short app_version;
	unsigned short crc;
	
} bootldrinfo_t;
const bootldrinfo_t bootlodrinfo __attribute__ ((section (".bootldrinfo"))) = {DEVID, SWVERSIONMAJOR << 8 | SWVERSIONMINOR, 0x0000};

int main(void)
{
	
	/* Enable necessary Clock sources. 32MHz and 32kHz Oscillator */
	OSC.CTRL = OSC_RC32MEN_bm;
	/* Wait for Oscillators to become stable */
	while(!(OSC.STATUS & OSC_RC32MRDY_bm));

	/* Start Calibrating the 32MHz Clock using
	* the internal calibrated 32kHz Oscillator */
	OSC.DFLLCTRL = 0;
	DFLLRC32M.CTRL = DFLL_ENABLE_bm;	// <<------------
	/* Set as System Main Clock */
	CCP = CCP_IOREG_gc;
	CLK.CTRL = CLK_SCLKSEL_RC32M_gc;

	/* Lock Clock Settings until next Reset */
	CCP = CCP_IOREG_gc;
	CLK.LOCK = CLK_LOCK_bm;
	
	// USART data struct used in task
	
	
	// PD3 (TX) as output
	USARTPORT.DIRSET   = PIN3_bm;
	// PD2 (RX) as input
	USARTPORT.DIRCLR   = PIN2_bm;
	
	// Use USARTD0 and initialize buffers
	USART_InterruptDriver_Initialize(&USART_data, &USART, USART_DREINTLVL_LO_gc);
	
	// USARTD0, 8 Data bits, No Parity, 1 Stop bit
	USART_Format_Set(USART_data.usart, USART_CHSIZE_8BIT_gc,
	USART_PMODE_DISABLED_gc, false);
	
	// Enable RXC interrupt
	USART_RxdInterruptLevel_Set(USART_data.usart, USART_RXCINTLVL_LO_gc);
	
	// Set Baudrate to 9600 bps:
	// Use the default I/O clock frequency that is 2 MHz.
	// Do not use the baudrate scale factor
	//
	// Baudrate select = (1/(16*(((I/O clock frequency)/Baudrate)-1)
	//                 = 12
	
	USART_Baudrate_Set(&USART, 1047 , -6);

	/* Enable both RX and TX. */
	USART_Rx_Enable(USART_data.usart);
	USART_Tx_Enable(USART_data.usart);
	

	// Enable PMIC interrupt level low
	PMIC.CTRL |= PMIC_LOLVLEX_bm;

	// Enable global interrupts
	sei();
	
	USART_TXBuffer_PutByte(&USART_data, 'A');
	USART_TXBuffer_PutByte(&USART_data, 'P');
	USART_TXBuffer_PutByte(&USART_data, 'P');
	UART_newline(&USART_data);
	//	uart_puts_p(&USART_data,PSTR("B/L start"));
	
	
	PORTE.DIRSET = PIN3_bm;
	
	while(1)
	{
		PORTE.OUTTGL = PIN3_bm;
		_delay_ms(80);
	}
}





ISR( USARTC0_RXC_vect )
{
	USART_RXComplete( &USART_data );
}


//  Data register empty  interrupt service routine.
//  Calls the common data register empty complete handler with pointer to the
//  correct USART as argument.

ISR( USARTC0_DRE_vect )
{
	USART_DataRegEmpty( &USART_data );
	
}

