/*-------------------------------------------------------------------------/
/  Stand-alone MMC boot loader  R0.01
/--------------------------------------------------------------------------/
/

/
/ The following is existing copyright text from Chan's "pfboot" (some parts
/ of this text are old an no longer apply to this project):
/
/  Copyright (C) 2010, ChaN, all right reserved.
/
/ * This software is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/--------------------------------------------------------------------------/
/ Dec 6, 2010  R0.01  First release
/--------------------------------------------------------------------------/
/ This is a stand-alone MMC/SD boot loader for XmegaAVRs. It requires a 4KB
/ boot section for code, four GPIO pins for MMC/SD as shown in sch.jpg and
/ nothing else. To port the boot loader into your project, follow the
/ instruction described below.
/
/ 1. Setup the hardware. Attach a memory card socket to the any GPIO port
/    where you like. Select boot size at least 4KB for the boot loader with
/    BOOTSZ fuses and enable boot loader with BOOTRST fuse.
/
/ 2. Setup the software. Change the four port definitions in the configure.h.
/    Change MCU_TARGET and MCU_FREQ.Build the boot loader
/    and write it to the device with a programmer.
/
/ 3. Build the application program.
/
/ 4. Insert the card and turn the target power on. When the boot loader found
/    the application file, the file is written into the flash memory prior to
/    start the application program. On-board LED lights (if exist) during
/    the flash programming operation.
/
/-------------------------------------------------------------------------*/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <string.h>
#include <stdlib.h>
#include <util/crc16.h>

#include "integer.h"
#include "diskio.h"
#include "avr_compiler.h"
#include "pff.h"


#include "configure.h"

#if (F_CPU != 32000000UL)
#error F_CPU must match oscillator setting!
#endif // F_CPU




typedef struct
{
	unsigned long dev_id;
	unsigned short app_version;
	unsigned short crc;
} bootldrinfo_t;


void SP_EraseWriteApplicationPage(uint32_t address);
void SP_LoadFlashPage(const uint8_t * data);

bootldrinfo_t current_bootldrinfo;
bootldrinfo_t file_bootldrinfo;






FATFS Fatfs;				/* Petit-FatFs work area */
BYTE Buff[SPM_PAGESIZE];	/* Page data buffer */



//##########################################################################
//#                                           THE PARTY STARTS HERE ;-)
//##########################################################################

void __jMain(void) __attribute__ ((naked)) __attribute__ ((section (".init9")));
void __jMain(void)
{
	asm volatile ( ".set __stack, %0" :: "i" (RAMEND) );
	asm volatile ( "clr __zero_reg__" );        // gcc expcets this to be zero.
	asm volatile ( "rjmp main");                // start main()
}



int main(void) 
{
	
	DWORD fa;	/* Flash address */
	UINT br;	/* Bytes read */
	unsigned short flash_or_file_crc = 0xFFFF;
	
	unsigned long i;
	
	/* Enable necessary Clock sources. 32MHz and 32kHz Oscillator */
	OSC.CTRL = OSC_RC32MEN_bm;
	/* Wait for Oscillators to become stable */
	while(!(OSC.STATUS & OSC_RC32MRDY_bm));

	/* Start Calibrating the 32MHz Clock using
	* the internal calibrated 32kHz Oscillator */
	OSC.DFLLCTRL = 0;
	DFLLRC32M.CTRL = DFLL_ENABLE_bm;	// <<------------
	/* Set as System Main Clock */
	CCP = CCP_IOREG_gc;
	CLK.CTRL = CLK_SCLKSEL_RC32M_gc;

	/* Lock Clock Settings until next Reset */
	CCP = CCP_IOREG_gc;
	CLK.LOCK = CLK_LOCK_bm;
	
	//CCP = CCP_IOREG_gc;      // unlock IVSEL
	//PMIC.CTRL |= PMIC_IVSEL_bm;
	
	void (*funcptr)( void ) = 0x0000; // Set up function pointer to RESET vector.
	
	LED_PORT.DIRSET = LED_BIT;
	LED_PORT.OUTSET = LED_BIT;
	
	
	
	SD_PORT.DIRSET = CS_PIN;
	SD_PORT.PIN4CTRL = PORT_OPC_WIREDANDPULL_gc;
	SD_PORT.OUTSET = CS_PIN;
	
	

	
	memcpy_P(&current_bootldrinfo, (uint8_t*) APP_SECTION_END  - sizeof(bootldrinfo_t) + 1, sizeof(bootldrinfo_t));		
	if(current_bootldrinfo.app_version == 0xFFFF)		//if application not flashed jet
	current_bootldrinfo.app_version = 0;
	
	
	for (i=0; i<=APP_SECTION_END; i++)
	flash_or_file_crc = _crc_ccitt_update(flash_or_file_crc, pgm_read_byte_far(i));
	
	if(flash_or_file_crc)										//if application is damaged
	current_bootldrinfo.app_version = 0;
	
	flash_or_file_crc=0xffff;
	
	pf_mount(&Fatfs);	/* Initialize file system */
	
	if (pf_open("APP.BIN") == FR_OK) /* Open application file */
	{
		
		for (fa = 0; fa < APP_SECTION_END; fa += SPM_PAGESIZE) /* check all application pages */
		{
			
			memset(Buff, 0xFF, SPM_PAGESIZE);	/* Clear buffer */
			pf_read(Buff, SPM_PAGESIZE, &br);	/* Load a page data */
			if (br)
			{
				
				for (i = 0; i < SPM_PAGESIZE; i++)
				flash_or_file_crc = _crc_ccitt_update(flash_or_file_crc, Buff[i]);				//crc of the file
				
			}
		}
		memcpy(&file_bootldrinfo, Buff  + (SPM_PAGESIZE -sizeof(bootldrinfo_t)) , sizeof(bootldrinfo_t));
		
	}
	
	
	if (file_bootldrinfo.dev_id == DEVID && ((current_bootldrinfo.app_version < file_bootldrinfo.app_version) || (file_bootldrinfo.app_version == 0)) && (flash_or_file_crc==0)  )// right Device, newer Version or Zero Version and crc of file ok
	{
		if (pf_open("APP.BIN") == FR_OK) /* Open application file */
		{

			for (fa = 0; fa < APP_SECTION_END; fa += SPM_PAGESIZE) /* Update all application pages */
			{
				
				memset(Buff, 0xFF, SPM_PAGESIZE);	/* Clear buffer */
				pf_read(Buff, SPM_PAGESIZE, &br);	/* Load a page data */
				if (br)
				{
					
					SP_LoadFlashPage(Buff);
					SP_EraseWriteApplicationPage(fa);
					
					while(NVM.STATUS & NVM_FBUSY_bm );		//wait while flash not ready
					if((fa/SPM_PAGESIZE%2)==0)
					LED_PORT.OUTTGL = LED_BIT;
				}
			}
		}
	}
	
	flash_or_file_crc=0xffff;
	for (i=0; i<=APP_SECTION_END; i++)
	flash_or_file_crc = _crc_ccitt_update(flash_or_file_crc, pgm_read_byte_far(i));
	
	
	if (flash_or_file_crc==0)		/* Start application if exist */
	{
		funcptr(); // Jump to Reset vector 0x0000 in Application Section.
	}
	
	LED_PORT.OUTSET|= LED_BIT;
	
	while(1)		//no application
	{
		;		
	}
}


