/*-----------------------------------------------------------------------/
/  PFF - Generic low level disk control module            (C)ChaN, 2010
/------------------------------------------------------------------------/
/
/  Copyright (C) 2010, ChaN, all right reserved.
/
/ * This software is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/------------------------------------------------------------------------/
/ * Dec 6, 2010  First release
*/
#include <util/delay.h>
//#include "pff.h"
#include "diskio.h"
#include "avr_compiler.h"
#include <avr/io.h>
#include "configure.h"


#define SELECT()	SD_PORT.OUTCLR = CS_PIN	/* CS = L */







/*--------------------------------------------------------------------------

Module Private Functions

---------------------------------------------------------------------------*/

/* Definitions for MMC/SDC command */
#define CMD0	(0x40+0)	/* GO_IDLE_STATE */
#define CMD1	(0x40+1)	/* SEND_OP_COND (MMC) */
#define	ACMD41	(0xC0+41)	/* SEND_OP_COND (SDC) */
#define CMD8	(0x40+8)	/* SEND_IF_COND */
#define CMD16	(0x40+16)	/* SET_BLOCKLEN */
#define CMD17	(0x40+17)	/* READ_SINGLE_BLOCK */
#define CMD24	(0x40+24)	/* WRITE_BLOCK */
#define CMD55	(0x40+55)	/* APP_CMD */
#define CMD58	(0x40+58)	/* READ_OCR */


/* Card type flags (CardType) */
#define CT_MMC				0x01	/* MMC ver 3 */
#define CT_SD1				0x02	/* SD ver 1 */
#define CT_SD2				0x04	/* SD ver 2 */
#define CT_BLOCK			0x08	/* Block addressing */

BYTE CardType;


void init_spi (void)
{
	SD_PORT.DIRSET = CS_PIN;
	SD_PORT.PIN4CTRL = PORT_OPC_WIREDANDPULL_gc;
	SD_PORT.OUTSET = CS_PIN;
	SD_PORT.DIRSET = DI_PIN;
	SD_PORT.DIRSET = SCLK_PIN;
	SD_PORT.DIRCLR = DO_PIN; /* Pin 6 as input */
	SD_PORT.PIN6CTRL = PORT_OPC_PULLUP_gc; /* PIN6 WIRED AND PULLup */
	
	/* initialize SPI with lowest frequency; max. 400kHz during identification mode of card */

	SD_SPI.CTRL =  SPI_MODE_3_gc | SPI_PRESCALER_DIV128_gc  | SPI_ENABLE_bm | SPI_MASTER_bm; //250 kHz
	SD_SPI.INTCTRL = SPI_INTLVL_OFF_gc;
	SD_SPI.STATUS = 0;
}




/*-----------------------------------------------------------------------*/
/* Transmit a byte to MMC via SPI  (Platform dependent)                  */
/*-----------------------------------------------------------------------*/
void xmit_spi(BYTE dat)
{
	SD_SPI.DATA = dat; 	//Sendet ein Byte
	while(!(SD_SPI.STATUS & (1<<SPI_IF_bp))); //Wartet bis Byte gesendet wurde
	_delay_us(1);
}

void xmit_FF(void)
{
	xmit_spi(0xFF);
}

void DESELECT(void)
{	
	SD_PORT.OUTSET = CS_PIN;	/* CS = H */
	xmit_FF();
}
/*-----------------------------------------------------------------------*/
/* Receive a byte from MMC via SPI  (Platform dependent)                 */
/*-----------------------------------------------------------------------*/
BYTE rcv_spi (void)
{
	BYTE Byte = 0;
	SD_SPI.DATA = 0xff;
	while(!(SD_SPI.STATUS & (1<<SPI_IF_bp))){};
	Byte = SD_SPI.DATA;
	_delay_us(1);
	return (Byte);
	
	
	
}


static void	FCLK_FAST(void)
{
	//switch to full speed (1 MHz)
	SD_SPI.CTRL =  SPI_MODE_3_gc | SPI_PRESCALER_DIV16_gc  | SPI_ENABLE_bm | SPI_MASTER_bm; //250 kHz
	
}

/*-----------------------------------------------------------------------*/
/* Send a command packet to MMC                                          */
/*-----------------------------------------------------------------------*/

static
BYTE send_cmd (
BYTE cmd,		/* 1st byte (Start + Index) */
DWORD arg		/* Argument (32 bits) */
)
{
	BYTE n, res;


	if (cmd & 0x80) {	/* ACMD<n> is the command sequense of CMD55-CMD<n> */
		cmd &= 0x7F;
		res = send_cmd(CMD55, 0);
		xmit_spi(0xFF);
		if (res > 1) return res;
	}

	/* Select the card */
	SELECT();

	/* Send a command packet */
	xmit_spi(cmd);						/* Start + Command index */
	xmit_spi((BYTE)(arg >> 24));		/* Argument[31..24] */
	xmit_spi((BYTE)(arg >> 16));		/* Argument[23..16] */
	xmit_spi((BYTE)(arg >> 8));			/* Argument[15..8] */
	xmit_spi((BYTE)arg);				/* Argument[7..0] */
	n = 0x01;							/* Dummy CRC + Stop */
	if (cmd == CMD0) n = 0x95;			/* Valid CRC for CMD0(0) */
	if (cmd == CMD8) n = 0x87;			/* Valid CRC for CMD8(0x1AA) */
	xmit_spi(n);

	/* Receive a command response */
	n = 10;								/* Wait for a valid response in timeout of 10 attempts */
	do {
		res = rcv_spi();
	} while ((res & 0x80) && --n);
	//xmit_spi(0xFF);

	return res;			/* Return with the response value */
}



/*--------------------------------------------------------------------------

Public Functions

---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Initialize Disk Drive                                                 */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (void)
{
	BYTE n, cmd, ty, ocr[4];
	UINT tmr;


	init_spi();							/* Initialize ports to control MMC */
	_delay_ms(10);	/* 10ms delay */
	DESELECT();
	
	for (n = 10; n; n--) 	/* 80 Dummy clocks with CS=H */
	xmit_FF();
	
	ty = 0;
	if (send_cmd(CMD0, 0) == 1) /* Enter Idle state */
	{			
		xmit_FF();
		if (send_cmd(CMD8, 0x1AA) == 1) /* SDv2 */
		{	
			for (n = 0; n < 4; n++) ocr[n] = rcv_spi();		/* Get trailing return value of R7 resp */
			xmit_FF();
			if (ocr[2] == 0x01 && ocr[3] == 0xAA) 
			{			/* The card can work at vdd range of 2.7-3.6V */
				for (tmr = 10000; tmr && send_cmd(ACMD41, 1UL << 30); tmr--)
				{
					 xmit_FF();
					 _delay_us(100);
				}	/* Wait for leaving idle state (ACMD41 with HCS bit) */
				xmit_FF();
				if (tmr && send_cmd(CMD58, 1UL << 30) == 0) /* Check CCS bit in the OCR */
				{		
					for (n = 0; n < 4; n++) ocr[n] = rcv_spi();
					xmit_FF();
					ty = (ocr[0] & 0x40) ? CT_SD2 | CT_BLOCK : CT_SD2;	/* SDv2 (HC or SC) */
				}
			}
		} 
		else 
		{							/* SDv1 or MMCv3 */
			if (send_cmd(ACMD41, 0) <= 1) 	
			{
				ty = CT_SD1; cmd = ACMD41;	/* SDv1 */
			} 
			else 
			{
				ty = CT_MMC; cmd = CMD1;	/* MMCv3 */
			}
			xmit_FF();
			for (tmr = 10000; tmr && send_cmd(cmd, 0); tmr--)
			{
				 xmit_FF();
				 _delay_us(100);	/* Wait for leaving idle state */
			}
			if (!tmr || send_cmd(CMD16, 512) != 0)			/* Set R/W block length to 512 */
			{
				ty = 0;
				xmit_FF();
			}
			
		}
	}
	CardType = ty;
	DESELECT();
	FCLK_FAST();

	return ty ? 0 : STA_NOINIT;
}

/*-----------------------------------------------------------------------*/
/* Read partial sector                                                   */
/*-----------------------------------------------------------------------*/

DRESULT disk_readp (
BYTE *buff,		/* Pointer to the read buffer (NULL:Read bytes are forwarded to the stream) */
DWORD lba,		/* Sector number (LBA) */
UINT ofs,		/* Byte offset to read from (0..511) */
UINT cnt		/* Number of bytes to read (ofs + cnt mus be <= 512) */
)
{
	DRESULT res;
	BYTE rc;
	UINT bc;


	if (!(CardType & CT_BLOCK)) lba *= 512;		/* Convert to byte address if needed */

	res = RES_ERROR;
	if (send_cmd(CMD17, lba) == 0) {		/* READ_SINGLE_BLOCK */

		bc = 40000;
		do {							/* Wait for data packet */
			rc = rcv_spi();
		} while (rc == 0xFF && --bc);

		if (rc == 0xFE) {				/* A data packet arrived */
			bc = 514 - ofs - cnt;

			/* Skip leading bytes */
			if (ofs) {
				do rcv_spi(); while (--ofs);
			}

			/* Receive a part of the sector */
			do {
				*buff++ = rcv_spi();
			} while (--cnt);

			/* Skip trailing bytes and CRC */
			do rcv_spi(); while (--bc);

			res = RES_OK;
		}
	}

	DESELECT();

	return res;
}


